@@ -12,9 +12,10 @@
 #include <stdlib.h>
 #include <sys/timeb.h>
 #include <time.h>
+#include <sys/time.h>
 
 #define PROGRAM		"Wardrive"
-#define VERSION		"v2.1"
+#define VERSION		"v2.3"
 #define AUTHOR		"van Hauser / THC <vh@reptile.rug.ac.be>"
 #define RESSOURCE	"http://www.thehackerschoice.com"
 
@@ -37,6 +38,7 @@
 #define OUTPUT_FILE_HEADER "# WARDRIVE by van Hauser / THC\n# DATE     TIME     LATITUDE       LONGITUDE  STATUS LINK LEVEL NOISE NWID CRYPT MISC\n"
 
 #define MAXLEN 256
+#define MAXGPSLEN 1024
 
 typedef struct {
    int    Hours, Minutes;
@@ -89,21 +91,45 @@ PROGRAM, RESSOURCE);
     exit(-1);
 }
 
+void seriell_clean_up (int ignore) {
+    tcsetattr(fd_serport, TCSANOW, &serport_opts_orig);
+    close(fd_serport);
+    (void) unlink(lockfile);
+    if (! ignore)
+        exit(-1);
+    fd_serport = -1;
+    return;
+}
+
 void clean_up (int sig) {
-    if (pid > 0)
-        kill(pid, SIGTERM);
-    if (pid == 0 && fd_serport > 0) {
-        tcsetattr(fd_serport, TCSANOW, &serport_opts_orig);
-        close(fd_serport);
-        (void) unlink(lockfile);
+    if (pid != 0) {
+        if (pid > 0)
+            kill(pid, SIGTERM);
+        if (g != NULL)
+            fclose(g);
+        seriell_clean_up(1);
     }
-    if (pid > 0 && g != NULL)
-        fclose(g);
     if (sig == -1)
         exit(-1);
     exit(0);
 }
 
+char *concat(char *old, char *sep, char *add) {
+    char *ptr;
+    ptr = malloc(strlen(old) + strlen(sep) + strlen(add) + 1);
+    strcpy(ptr, old);
+    strcat(ptr, sep);
+    strcat(ptr, add);
+    return ptr;
+}
+
+int str2int(char *str, int no) {
+    char buf[no + 1];
+    buf[no] = 0;
+    memcpy(buf, str, no);
+    return atoi(buf);
+}
+
 int get_parameter(int ppos, char *value, int maxlen, char *msg) {
     int   idx = 0;
     char *cur = NULL;
@@ -129,7 +155,7 @@ int get_parameter(int ppos, char *value, int maxlen, char *msg) {
 
 int main(int argc, char *argv[]) {
     FILE *f, *iw;
-    char buf[MAXLEN], buf2[MAXLEN], pos[MAXLEN], pos2[MAXLEN], *ptr, *ptr2, datetime[24];
+    char buf[MAXLEN], buf2[MAXGPSLEN], pos[MAXLEN], *ptr, *ptr2, datetime[24];
     char *iwcfg = NULL, *ssid = NULL, *ap = NULL;
     POS coords, *position = &coords;
     int res, idx = 0, len, link_away = -1;
@@ -140,7 +166,6 @@ int main(int argc, char *argv[]) {
     time_t epoch;
     int socks[2];
     char location[MAXLEN] = "00:00:00.0000? 00:00:00.0000?";
-// XXX TODO: bug ... 00:00 after some time position is lost :-(
 
     device        = DEFAULT_INTERFACE;
     serport       = DEFAULT_SERPORT;
@@ -207,36 +232,17 @@ int main(int argc, char *argv[]) {
             ptr2 = malloc(strlen("PATH=."));
             strcpy(ptr2, "PATH=.");
         } else {
-            ptr2 = malloc(strlen(ptr) + strlen("PATH=") + 3);
-            strcpy(ptr2, "PATH=");
-            strcat(ptr2, ptr);
-            strcat(ptr2, ":.");
+            ptr2 = concat("PATH=", ptr, ":.");
         }
         putenv(ptr2);
         ptr = ptr2 = NULL;
     }
-    if (init != NULL) {
-    	ptr = malloc(strlen(init) + strlen(device) + 2);
-    	strcpy(ptr, init);
-    	strcat(ptr, " ");
-    	strcat(ptr, device);
-    	init = ptr;
-    	ptr = NULL;
-    }
-    if (reset != NULL) {
-    	ptr = malloc(strlen(reset) + strlen(device) + 2);
-    	strcpy(ptr, reset);
-    	strcat(ptr, " ");
-    	strcat(ptr, device);
-    	reset = ptr;
-    	ptr = NULL;
-    }
-    if (winfo) {
-        iwcfg = malloc(strlen("iwconfig") + strlen(device) + 2);
-        strcpy(iwcfg, "iwconfig");
-        strcat(iwcfg, " ");
-        strcat(iwcfg, device);
-    }
+    if (init != NULL)
+        init = concat(init, " ", device);
+    if (reset != NULL)
+        reset = concat(reset, " ", device);
+    if (winfo)
+        iwcfg = concat("iwconfig", " ", device);
 
     if ((f = fopen(PROC_WAVELAN, "r")) == NULL) {
         fprintf(stderr, "%s: Can not open %s for reading.\n", PROGRAM, PROC_WAVELAN);
@@ -255,7 +261,7 @@ int main(int argc, char *argv[]) {
     fclose(f);
 
     if (init != NULL) {
-        if (debug) printf("DEBUG RUNNING %s\n", init);
+        if (debug) printf("DEBUG: RUNNING %s\n", init);
         if (system(init) != 0)
             fprintf(stderr, "%s: Can not find init script %s\n", PROGRAM, init);
     }
@@ -300,33 +306,31 @@ int main(int argc, char *argv[]) {
         serport_opts.c_cc[VMIN] = 0;               // set min read characters if 0, VTIME takes over
         serport_opts.c_cc[VTIME] = V_TIME;         // wait V_TIME ms for character
         if (ignore_errors <= 1 && (res = cfsetospeed(&serport_opts, B4800)) < 0) {
-            unlink(lockfile);
-            close (fd_serport);
-            if (! ignore_errors) {
-                fprintf(stderr, "%s: Unable to configure seriell port %s, use option -G to disable GPS.\n", PROGRAM, serport);
-                exit(-1);
-            }
+            fprintf(stderr, "%s: Unable to configure seriell port %s of GPS device.\n", PROGRAM, serport);
+            seriell_clean_up(ignore_errors);
             ignore_errors++;
         }
         if (ignore_errors <= 1 && tcsetattr(fd_serport,TCSANOW, &serport_opts) < 0) {
-            unlink(lockfile);
-            close (fd_serport);
-            if (! ignore_errors) {
-                fprintf(stderr, "%s: Unable to configure seriell port %s, use option -G to disable GPS.\n", PROGRAM, serport);
-                exit(-1);
-            }
+            fprintf(stderr, "%s: Unable to configure seriell port %s of GPS device.\n", PROGRAM, serport);
+            seriell_clean_up(ignore_errors);
             ignore_errors++;
         }
-        if (ignore_errors < 2) {
+        if (ignore_errors <= 1) {
+            int errors = 0;
             buf[0] = ' ';
-            while (buf[0] != '\r')
-                read (fd_serport, &buf[0], 1);
+            while (buf[0] != '\r' && errors < 3)
+                if (read(fd_serport, &buf[0], 1) < 1)
+                    errors++;
+	    if (errors >= 3) {
+	        fprintf(stderr, "%s: No response from GPS on seriell port %s.\n", PROGRAM, serport);
+	        seriell_clean_up(ignore_errors);
+	        ignore_errors++;
+	    }
         } else
             fd_serport = -1;
     }
     if (ignore_errors > 1)
         fprintf(stderr, "%s: GPS could not be configured, disabled support and still running ...\n", PROGRAM);
-
     if ((res = open(file, O_RDWR | O_APPEND)) < 0) {
         if ((res  = open(file, O_RDWR | O_APPEND | O_CREAT | O_EXCL, 0600)) < 0) {
             fprintf(stderr, "%s: Unable to create file %s\n", PROGRAM, file);
@@ -339,7 +343,8 @@ int main(int argc, char *argv[]) {
     f = fopen(file, "a");
     fclose(g);
     g = f;
-//    setvbuf(g, NULL, _IONBF, 0);
+    if (link_save > 0)
+        setvbuf(g, NULL, _IONBF, 0);
 
     if (ignore_errors < 2) {
         if (socketpair(PF_UNIX, SOCK_STREAM, 0, socks) != 0) {
@@ -348,30 +353,43 @@ int main(int argc, char *argv[]) {
         }
 
         if ((pid = fork()) == 0) {
-            char reading;
+            char reading, buf3[48], *pptr = NULL, *tptr = NULL;
+            int lost = 1;
 
             memset(&coords, 0, sizeof(coords));
             fcntl(socks[0], F_SETFL, O_NONBLOCK);
             while(1) {
                 res = 0;
+                memset(buf2, 0, MAXGPSLEN);
 
+                while(read(fd_serport, &buf2, MAXGPSLEN - 1) <= 0);
+
+                while (res == 0) {
+ 		    if ((pptr = rindex(buf2, '\n')) != NULL && (pptr = strstr(buf2, "GLL")) != NULL && *(pptr -= 3) == '$') {
+		        if ((tptr = index(pptr, '\r')) != NULL)
+		            *tptr = 0;
+		        if ((tptr = index(pptr, '\n')) != NULL)
+		            *tptr = 0;
+		        res = 1;
+		        if (debug && verbose) printf("DEBUG: GPS MSG - %s\n", pptr);
+		    }
+		    if (pptr == NULL)
+		        res = -1;
+		}
+
+/* This is the old function which resulted in a high backlog of GPS msgs
+		len = 0;
                 while (!res) {
-                    memset(buf2, 0, MAXLEN);
-                    len = 0;
-                    while (! res) {
-                        if (read(fd_serport, &reading, 1) == 1) {
-                            if (reading != '\r') {
-                                buf2[len] = reading;
-                                len++;
-                            }
-                            if (reading == '\n')
-                                res = 1;
+                    if (read(fd_serport, &reading, 1) == 1) {
+                        if (reading != '\r') {
+                            buf2[len] = reading;
+                        }
+                        if (reading == '\n')
+                            res = 1;
                         }
-                    }
                 }
-
-                if (debug) printf("DEBUG GPS MSG: %d - %s", len, buf2);
-
+                len = strlen(buf2);
+                if (debug && verbose) printf("DEBUG: GPS MSG length %d - %s", len, buf2);
                 len = strlen(NMEA_POS_STRING);
                 for (idx = 0; idx < len; idx++) {
                    if (NMEA_POS_STRING[idx] == '?')
@@ -380,76 +398,82 @@ int main(int argc, char *argv[]) {
                        res = 0;
                    }
                 }
+                pptr = buf2;
+*/
 
-		if (debug && res && strstr(buf2, ",,,") != NULL)
-		    printf("DEBUG: EMPTY GPS POS MSG FOUND: %s\n", buf2);
+		if (res == 1 && strstr(pptr, ",,,") != NULL) {
+		    res = 0;
+		    if (debug && verbose == 0)
+ 		        printf("DEBUG: EMPTY GPS POS MSG FOUND: %s\n", pptr);
+ 		    else if (verbose) {
+ 		        if (lost == 0) {
+ 		            lost = 1;
+ 		            printf("%s: Warning - GPS lost position!\n", prg);
+ 		        }
+ 		    }
+		}
+		if (res == 1 && index(pptr + 2, '$') != NULL) {
+		    res = 0;
+		    if (debug) {
+		        printf("DEBUG: FUCKED UP GPS POS MSG FOUND: %s\n", pptr);
+		    }
+		}
 
-		if (res == 1 && strstr(buf2, ",,,") == NULL) {
-//                    if (debug) printf("DEBUG: GPS MSG FOUND\n");
-
-                    get_parameter(1, pos, 25, buf2);
-                    pos2[0] = pos[0];
-                    pos2[1] = pos[1];
-                    pos2[3] = '\0';
-                    position->Latitude.Hours = atoi(pos2);
+		if (res == 1) {
+		    lost = 0;
+                    get_parameter(1, pos, 25, pptr);
+                    position->Latitude.Hours = str2int(pos, 2);
                     len = strlen(pos) + 1;
                     for (idx = 2; idx < len; idx++)
                          pos[idx - 2] = pos[idx];
-                    pos2[0] = pos[0];
-                    pos2[1] = pos[1];
-                    position->Latitude.Minutes = atoi(pos2);
+                    position->Latitude.Minutes = str2int(pos, 2);
                     len = strlen(pos) + 1;
                     for (idx = 3; idx < len; idx++)
                         pos[idx - 3] = pos[idx];
                     position->Latitude.Seconds = ((atoi(pos) * 1.0) / 10000) * 60;
 
-                    get_parameter(2, pos, 25, buf2);
+                    get_parameter(2, pos, 25, pptr);
                     position->Latitude.Dir = pos[0];
 
-                    get_parameter(3, pos, 25, buf2);
-                    pos2[0] = pos[0];
-                    pos2[1] = pos[1];
-                    pos2[2] = pos[2];
-                    pos2[3] = '\0';
-                    position->Longitude.Hours = atoi(pos2);
+                    get_parameter(3, pos, 25, pptr);
+		    position->Longitude.Hours = str2int(pos,3);
                     len = strlen (pos) + 1;
                     for (idx = 3; idx < len; idx++)
                     pos[idx - 3] = pos[idx];
-                    pos2[0] = pos[0];
-                    pos2[1] = pos[1];
-                    pos2[2] = '\0';
-                    position->Longitude.Minutes = atoi(pos2);
+                    position->Longitude.Minutes = str2int(pos, 2);
                     len = strlen (pos) + 1;
                     for (idx = 3; idx < len; idx++)
                     pos[idx - 3] = pos[idx];
                     position->Longitude.Seconds = ((atoi(pos) * 1.0) / 10000) * 60;
  
-                    get_parameter(4, pos, 25, buf2);
+                    get_parameter(4, pos, 25, pptr);
                     position->Longitude.Dir = pos[0];
 
                     if (position->Latitude.Dir != 'N' && position->Latitude.Dir != 'S')
                         position->Latitude.Dir = '?';
                     if (position->Longitude.Dir != 'E' && position->Longitude.Dir != 'W')
                         position->Longitude.Dir = '?';
-            
-/*                    if (debug) {
-                        printf ("DEBUG LATITUDE  - Hours: %02d Minutes: %02d Seconds: %02.04f %c\n",
-                          position->Latitude.Hours, position->Latitude.Minutes,
-                          position->Latitude.Seconds, position->Latitude.Dir);
-                        printf ("DEBUG LONGITUDE - Hours: %02d Minutes: %02d Seconds: %02.04f %c\n",
-                          position->Longitude.Hours, position->Longitude.Minutes,
-                          position->Longitude.Seconds, position->Longitude.Dir);
+                    
+// this is a temporary debugging function to find out why it sometimes fucks up
+/*
+                    if (position->Longitude.Dir == '?' || position->Latitude.Dir == '?') {
+                        FILE *df;
+			if ((df=fopen("./.wardrive.DEBUG", "a")) != NULL) {
+			    fprintf(df, "DEBUG\nGPS: %spos: %02d:%02d:%02.04f%c %02d:%02d:%02.04f%c\n\n",pptr,
+                              position->Latitude.Hours, position->Latitude.Minutes, position->Latitude.Seconds, position->Latitude.Dir,
+                              position->Longitude.Hours, position->Longitude.Minutes, position->Longitude.Seconds, position->Longitude.Dir);
+			}
                     }
-*/                }
+*/
+                }
 
 		if (read(socks[1], &reading, 1) > 0) {
-		    char buf3[48];
-		    snprintf(buf3, sizeof(buf3), "%02d:%02d:%02.04f%c %02d:%02d:%02.04f%c\n",
-                        position->Latitude.Hours, position->Latitude.Minutes, position->Latitude.Seconds, position->Latitude.Dir,
-                        position->Longitude.Hours, position->Longitude.Minutes, position->Longitude.Seconds, position->Longitude.Dir);
+		    if (position->Longitude.Dir != '?' && position->Latitude.Dir != '?')
+ 		        snprintf(buf3, sizeof(buf3), "%02d:%02d:%02.04f%c %02d:%02d:%02.04f%c\n",
+                          position->Latitude.Hours, position->Latitude.Minutes, position->Latitude.Seconds, position->Latitude.Dir,
+                          position->Longitude.Hours, position->Longitude.Minutes, position->Longitude.Seconds, position->Longitude.Dir);
                     write(socks[1], &buf3, strlen(buf3));
-//		    if (debug) printf("DEBUG: REQUEST FOR POSITION -> %s\n", buf3);
-		}
+                }
             }
         }
         if (pid < 0) {
@@ -484,7 +508,6 @@ int main(int argc, char *argv[]) {
             clean_up(-1);
         }
         fclose(f);
-//        if (debug) printf("DEBUG PROC_WAVELAN: %s", buf);
         ptr = strstr(buf, device) + strlen(device) + 2;
         while (*ptr == 0 || *ptr == ' ' || *ptr == '\t') ptr++;
 
@@ -534,7 +557,6 @@ int main(int argc, char *argv[]) {
         while (*ptr2 != '\r' && *ptr2 != '\n' && *ptr2 != 0 && *ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '.') ptr2++;
         if (ptr2 != NULL) *ptr2 = 0;
         discard_misc = atoi(ptr);
-//        if (debug) printf("DEBUG WVLAN STATS status: %d, link: %d, level: %d, noise: %d, d-nwid: %d, d-crypt: %d, d-misc: %d\n", status, link, level, noise, discard_nwid, discard_crypt, discard_misc);
 
 /* gps */
         if (pid > 0) {
@@ -545,27 +567,23 @@ int main(int argc, char *argv[]) {
 	    location[end] = '\0';
 	    if (strlen(location) > 0 && location[strlen(location)-1] == '\n')
 	        location[strlen(location)-1] = '\0';
-	    if (strlen(location) == 0 || strstr(location, "00:00:0") != NULL)
+	    if (strlen(location) == 0 || strstr(location, "00:00:0") != NULL || index(location, '?') != NULL)
 	        strcpy(location, "00:00:00.0000? 00:00:00.0000?");
         }
 
 /* save data */
-        if (link >= link_save) {
-	    time(&epoch);
-	    the_time = localtime(&epoch);
-	    strftime(datetime, sizeof(datetime), "%Y-%m-%d %H:%M:%S", the_time);
+	time(&epoch);
+	the_time = localtime(&epoch);
+	strftime(datetime, sizeof(datetime), "%Y-%m-%d %H:%M:%S", the_time);
+
+        if (link >= link_save)
             fprintf(g, "%s %s %d %d %d %d %d %d %d\n",
 		datetime, location,
 		status, link, level, noise, discard_nwid, discard_crypt, discard_misc);
-	}
-        if ((link >= link_save && verbose) || debug) {
-	    time(&epoch);
-	    the_time = localtime(&epoch);
-	    strftime(datetime, sizeof(datetime), "%Y-%m-%d %H:%M:%S", the_time);
+        if ((link >= link_save && verbose) || debug)
             printf("%s %s %d %d %d %d %d %d %d\n",
 		datetime, location,
 		status, link, level, noise, discard_nwid, discard_crypt, discard_misc);
-	}
 
 /* beeping */
         ftime(&timing);
@@ -578,15 +596,17 @@ int main(int argc, char *argv[]) {
 
         if (link_away == 4 && reset != NULL) {
 	    link_away = -1;
-	    if (debug) printf("DEBUG RUNNING %s\n", reset);
+	    if (debug) printf("DEBUG: RUNNING %s\n", reset);
 	    if (system(reset) != 0) {
 	        fprintf(stderr, "%s: Can not find reset script %s, disabling -R option\n", PROGRAM, reset);
 	        reset = NULL;
 	    }
         }
 
+/* get SSID and MAC from found wavelan */
         if (link > 0) {
             if (winfo && link_away == -1) {
+                ap = ssid = NULL;
                 iw = popen(iwcfg, "r");
                 (void) wait3(NULL, WNOHANG, NULL);
                 while (fgets(buf, sizeof(buf), iw) != NULL) {
@@ -602,9 +622,6 @@ int main(int argc, char *argv[]) {
                     }
                 }
                 if (ssid != NULL || ap != NULL) {
-  	            time(&epoch);
-	            the_time = localtime(&epoch);
-	            strftime(datetime, sizeof(datetime), "%Y-%m-%d %H:%M:%S", the_time);
                     fprintf(g, "%s %s WINFO - %s %s",
  		        datetime, location, ssid, ap);
                     if (verbose) printf("%s %s WINFO - %s %s",
@@ -618,7 +635,7 @@ int main(int argc, char *argv[]) {
             }
             link_away = 0;
 	} else
-	    if (link_away >= 0) link_away++;
+	    if (link_away >= 0 && link_away < 255) link_away++;
 
 /* wait for next check */
         tmp1 = timing.time;
@@ -628,7 +645,7 @@ int main(int argc, char *argv[]) {
         if (save_interval > timer) {
             struct timeval tv;
             timer = save_interval - timer;
-//            if (debug) printf("DEBUG WAITING FOR %lu milliseconds\n", timer);
+//            if (debug && verbose) printf("DEBUG: WAITING FOR %lu milliseconds\n", timer);
             tv.tv_sec = timer / 1000;
             tv.tv_usec = (timer % 1000) * 1000;
             select(0, NULL, NULL, NULL, &tv);
